@@ -77,6 +77,8 @@ gint64 get_disc_size_dvdr   (int fd, int mmc_profile);
 int    get_read_write_speed (int fd, int *read_speed, int *write_speed);
 int    get_disc_status      (int fd, int *empty, int *is_rewritable, int *is_blank);
 
+#define ENABLE_NON_HAL_FALLBACK_CODE 1
+
 #ifndef INVALID_HANDLE
 #define INVALID_HANDLE (GINT_TO_POINTER(-1))
 #endif
@@ -112,6 +114,8 @@ struct NautilusBurnDrivePrivate {
 	gboolean                      monitor_enabled;
 	guint                         poll_id;
 
+	int                          *write_speeds;
+
 #ifdef USE_HAL
 	LibHalContext                *ctx;
 #endif
@@ -141,6 +145,24 @@ static GObjectClass *parent_class = NULL;
 
 G_DEFINE_TYPE(NautilusBurnDrive, nautilus_burn_drive, G_TYPE_OBJECT);
 
+/* fill in the write speeds as max_speed downto 1 if we
+ * are using a detection method that doesn't report all
+ * of the drive speeds
+ */
+static void
+fill_write_speeds (NautilusBurnDrive *drive)
+{
+	int max_speed;
+	int i;
+
+	max_speed = drive->max_speed_write;
+	drive->priv->write_speeds = g_new0 (int, max_speed + 1);
+
+	for (i = 0; i < max_speed; i++) {
+		drive->priv->write_speeds[i] = max_speed - i;
+	}
+}
+
 #ifdef __FreeBSD__
 
 #define get_ioctl_handle_fd(x) (((struct cam_device *)x)->fd)
@@ -443,6 +465,30 @@ nautilus_burn_drive_get_max_speed_read (NautilusBurnDrive *drive)
 	return speed;
 }
 
+/**
+ * nautilus_burn_drive_get_max_write_speeds:
+ * @drive: #NautilusBurnDrive
+ *
+ * Get the list of supported write speeds
+ *
+ * Returns: The write speeds of the drive, in device units, as a
+ * zero-terminated array of integers.  The array is sorted in
+ * descending order and has no duplicate entries.
+ *
+ * Since: 2.14
+ **/
+const int *
+nautilus_burn_drive_get_write_speeds (NautilusBurnDrive *drive)
+{
+	const int *speeds;
+
+	g_return_val_if_fail (drive != NULL, NULL);
+
+	speeds = drive->priv->write_speeds;
+
+	return speeds;
+}
+
 /**
  * nautilus_burn_drive_get_name_for_display:
  * @drive: #NautilusBurnDrive
@@ -909,6 +955,57 @@ nautilus_burn_drive_get_media_type (NautilusBurnDrive *drive)
 }
 
 #ifdef USE_HAL
+
+static char *
+hal_drive_get_first_child_udi (NautilusBurnDrive *drive)
+{
+	LibHalContext        *ctx;
+	char                **device_names;
+	int                   num_devices;
+	DBusError             error;
+	char                 *udi;
+
+	udi = NULL;
+
+	if (drive->priv->udi == NULL) {
+		return NULL;
+	}
+
+	ctx = drive->priv->ctx;
+
+	if (ctx == NULL) {
+		return NULL;
+	}
+
+	num_devices = -1;
+	dbus_error_init (&error);
+	device_names = libhal_manager_find_device_string_match (ctx, 
+								"info.parent",
+								drive->priv->udi,
+								&num_devices,
+								&error);
+
+	if (dbus_error_is_set (&error)) {
+		g_warning ("%s\n", error.message);
+		dbus_error_free (&error);
+
+		goto done;
+	}
+
+	if (num_devices <= 0) {
+		/*g_warning ("No HAL devices found for UDI '%s'", drive->priv->udi);*/
+
+		goto done;
+	}
+
+	udi = g_strdup (device_names [0]);
+
+ done:
+	libhal_free_string_array (device_names);
+
+	return udi;
+}
+
 static NautilusBurnMediaType
 nautilus_burn_drive_hal_get_media_type_full (NautilusBurnDrive *drive,
 					     gboolean          *is_rewritable,
@@ -916,129 +1013,106 @@ nautilus_burn_drive_hal_get_media_type_full (NautilusBurnDrive *drive,
 					     gboolean          *has_data,
 					     gboolean          *has_audio)
 {
+	LibHalContext        *ctx;
+	NautilusBurnMediaType type;
+	char                 *hal_type;
+	char                 *child_udi;
+
 	if (is_rewritable) *is_rewritable = FALSE;
 	if (is_blank) *is_blank = FALSE;
 	if (has_data) *has_data = FALSE;
 	if (has_audio) *has_audio = FALSE;
 
-	if (drive->priv != NULL && drive->priv->udi != NULL) {
-		LibHalContext        *ctx;
-		char                **device_names;
-		int                   num_devices = -1;
-		NautilusBurnMediaType type;
-		char                 *hal_type;
-		DBusError             error;
-		
-		ctx = drive->priv->ctx;
+	if (drive->priv->udi == NULL) {
+		goto fallback;
+	}
 
-		dbus_error_init (&error);
-		if (ctx != NULL) {
-			device_names = libhal_manager_find_device_string_match (ctx, 
-										"info.parent",
-										drive->priv->udi,
-										&num_devices,
-										&error);
+	ctx = drive->priv->ctx;
 
-			if (dbus_error_is_set (&error)) {
-				g_warning ("%s\n", error.message);
-				dbus_error_free (&error);
-				libhal_free_string_array (device_names);
+	child_udi = hal_drive_get_first_child_udi (drive);
+	if (child_udi == NULL) {
+		goto fallback;
+	}
 
-				goto fallback;
-			}
+	if (is_rewritable)
+		*is_rewritable = libhal_device_get_property_bool (ctx, 
+								  child_udi,
+								  "volume.disc.is_rewritable",
+								  NULL);
 
-			if (num_devices <= 0) {
-				g_warning ("No HAL devices found for UDI '%s'", drive->priv->udi);
-				libhal_free_string_array (device_names);
+	if (is_blank)
+		*is_blank = libhal_device_get_property_bool (ctx, 
+							     child_udi,
+							     "volume.disc.is_blank",
+							     NULL);
 
-				goto fallback;
-			}
+	if (has_data)
+		*has_data = libhal_device_get_property_bool (ctx, 
+							     child_udi,
+							     "volume.disc.has_data",
+							     NULL);
 
-			/* just look at the first child */
-			if (libhal_device_get_property_bool (ctx, 
-							     device_names [0],
-							     "volume.is_mounted",
-							     NULL)) {
-				type = NAUTILUS_BURN_MEDIA_TYPE_BUSY;
-			} else {
-			
-				if (is_rewritable)
-					*is_rewritable = libhal_device_get_property_bool (ctx, 
-									 		  device_names [0],
-											  "volume.disc.is_rewritable",
-											  NULL);
+	if (has_audio)
+		*has_audio = libhal_device_get_property_bool (ctx, 
+							      child_udi,
+							      "volume.disc.has_audio",
+							      NULL);
 
-				if (is_blank)
-					*is_blank = libhal_device_get_property_bool (ctx, 
-										     device_names [0],
-										     "volume.disc.is_blank",
-										     NULL);
+	type = NAUTILUS_BURN_MEDIA_TYPE_BUSY;
+	hal_type = libhal_device_get_property_string (ctx, 
+						      child_udi,
+						      "volume.disc.type",
+						      NULL);
 
-				if (has_data)
-					*has_data = libhal_device_get_property_bool (ctx, 
-										     device_names [0],
-										     "volume.disc.has_data",
-										     NULL);
-
-				if (has_audio)
-					*has_audio = libhal_device_get_property_bool (ctx, 
-										      device_names [0],
-										      "volume.disc.has_audio",
-										      NULL);
-
-				type = NAUTILUS_BURN_MEDIA_TYPE_BUSY;
-				hal_type = libhal_device_get_property_string (ctx, 
-									      device_names [0],
-									      "volume.disc.type",
-									      NULL);
-
-				if (hal_type == NULL || strcmp (hal_type, "unknown") == 0) {
-					type = NAUTILUS_BURN_MEDIA_TYPE_UNKNOWN;
-				} else if (strcmp (hal_type, "cd_rom") == 0) {
-					type = NAUTILUS_BURN_MEDIA_TYPE_CD;
-				} else if (strcmp (hal_type, "cd_r") == 0) {
-					type = NAUTILUS_BURN_MEDIA_TYPE_CDR;
-				} else if (strcmp (hal_type, "cd_rw") == 0) {
-					type = NAUTILUS_BURN_MEDIA_TYPE_CDRW;
-				} else if (strcmp (hal_type, "dvd_rom") == 0) {
-					type = NAUTILUS_BURN_MEDIA_TYPE_DVD;
-				} else if (strcmp (hal_type, "dvd_r") == 0) {
-					type = NAUTILUS_BURN_MEDIA_TYPE_DVDR;
-				} else if (strcmp (hal_type, "dvd_ram") == 0) {
-					type = NAUTILUS_BURN_MEDIA_TYPE_DVD_RAM;
-				} else if (strcmp (hal_type, "dvd_rw") == 0) {
-					type = NAUTILUS_BURN_MEDIA_TYPE_DVDRW;
-				} else if (strcmp (hal_type, "dvd_plus_rw") == 0) {
-					type = NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_RW;
-				} else if (strcmp (hal_type, "dvd_plus_r") == 0) {
-					type = NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_R;
-				} else if (strcmp (hal_type, "dvd_plus_r_dl") == 0) {
-					type = NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_R_DL;
-				} else {
-					type = NAUTILUS_BURN_MEDIA_TYPE_UNKNOWN;
-				}
+	if (hal_type == NULL || strcmp (hal_type, "unknown") == 0) {
+		type = NAUTILUS_BURN_MEDIA_TYPE_UNKNOWN;
+	} else if (strcmp (hal_type, "cd_rom") == 0) {
+		type = NAUTILUS_BURN_MEDIA_TYPE_CD;
+	} else if (strcmp (hal_type, "cd_r") == 0) {
+		type = NAUTILUS_BURN_MEDIA_TYPE_CDR;
+	} else if (strcmp (hal_type, "cd_rw") == 0) {
+		type = NAUTILUS_BURN_MEDIA_TYPE_CDRW;
+	} else if (strcmp (hal_type, "dvd_rom") == 0) {
+		type = NAUTILUS_BURN_MEDIA_TYPE_DVD;
+	} else if (strcmp (hal_type, "dvd_r") == 0) {
+		type = NAUTILUS_BURN_MEDIA_TYPE_DVDR;
+	} else if (strcmp (hal_type, "dvd_ram") == 0) {
+		type = NAUTILUS_BURN_MEDIA_TYPE_DVD_RAM;
+	} else if (strcmp (hal_type, "dvd_rw") == 0) {
+		type = NAUTILUS_BURN_MEDIA_TYPE_DVDRW;
+	} else if (strcmp (hal_type, "dvd_plus_rw") == 0) {
+		type = NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_RW;
+	} else if (strcmp (hal_type, "dvd_plus_r") == 0) {
+		type = NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_R;
+	} else if (strcmp (hal_type, "dvd_plus_r_dl") == 0) {
+		type = NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_R_DL;
+	} else {
+		type = NAUTILUS_BURN_MEDIA_TYPE_UNKNOWN;
+	}
 				
-				if (hal_type != NULL)
-					libhal_free_string (hal_type);
-			}
+	if (hal_type != NULL) {
+		libhal_free_string (hal_type);
+	}
 
-			libhal_free_string_array (device_names);
+	g_free (child_udi);
 
-			/* if the type has been determined using HAL, return it.
-			   Otherwise, fall through and try the non-HAL approach. */
-			if (type != NAUTILUS_BURN_MEDIA_TYPE_UNKNOWN
-			    && type != NAUTILUS_BURN_MEDIA_TYPE_ERROR) {
-				return type;
-			}
-		}
+	/* if the type has been determined using HAL, return it.
+	   Otherwise, fall through and try the non-HAL approach. */
+	if (type != NAUTILUS_BURN_MEDIA_TYPE_UNKNOWN
+	    && type != NAUTILUS_BURN_MEDIA_TYPE_ERROR) {
+		return type;
 	}
 
  fallback:
+#ifdef ENABLE_NON_HAL_FALLBACK_CODE
 	return nautilus_burn_drive_get_media_type_from_path_full (drive->device,
 								  is_rewritable,
 								  is_blank,
 								  has_data,
 								  has_audio);
+#else
+	return NAUTILUS_BURN_MEDIA_TYPE_ERROR;
+#endif
 }
 #endif
 
@@ -1149,6 +1223,7 @@ nautilus_burn_drive_get_media_size_from_path (const char *device)
 	return size;
 }
 
+#ifdef ENABLE_NON_HAL_FALLBACK_CODE
 static gint64
 try_media_size_from_type (NautilusBurnDrive *drive)
 {
@@ -1219,6 +1294,56 @@ try_media_size_from_type (NautilusBurnDrive *drive)
 
 	return size;
 }
+#endif
+
+#ifdef USE_HAL
+static gint64
+nautilus_burn_drive_hal_get_media_size (NautilusBurnDrive *drive)
+{
+	gint64         size;
+	guint64        res;
+	LibHalContext *ctx;
+	char          *child_udi;
+	DBusError      error;
+
+	size = -1;
+
+	if (drive->priv->udi == NULL) {
+		goto done;
+	}
+		
+	ctx = drive->priv->ctx;
+
+	if (ctx == NULL) {
+		goto done;
+	}
+
+	child_udi = hal_drive_get_first_child_udi (drive);
+	if (child_udi == NULL) {
+		goto done;
+	}
+
+	dbus_error_init (&error);
+	res = libhal_device_get_property_uint64 (ctx, 
+						 child_udi,
+						 "volume.disc.capacity",
+						 &error);
+
+	g_free (child_udi);
+
+	if (dbus_error_is_set (&error)) {
+		g_warning ("%s\n", error.message);
+		dbus_error_free (&error);
+				
+		goto done;
+	}
+
+	size = (gint64) res;
+
+ done:
+	return size;
+}
+#endif
 
 /**
  * nautilus_burn_drive_get_media_size:
@@ -1241,10 +1366,17 @@ nautilus_burn_drive_get_media_size (NautilusBurnDrive *drive)
 	g_return_val_if_fail (drive != NULL, NAUTILUS_BURN_MEDIA_SIZE_UNKNOWN);
 
 #ifdef USE_HAL
-	size = nautilus_burn_drive_get_media_size_from_path (drive->device);
+	size = nautilus_burn_drive_hal_get_media_size (drive);
+
+#ifdef ENABLE_NON_HAL_FALLBACK_CODE
 	if (size < 0) {
-		size = try_media_size_from_type (drive);
+		size = nautilus_burn_drive_get_media_size_from_path (drive->device);
+		if (size < 0) {
+			size = try_media_size_from_type (drive);
+		}
 	}
+#endif
+
 #else
 	size = nautilus_burn_drive_get_media_size_from_path (drive->device);
 #endif
@@ -1252,6 +1384,77 @@ nautilus_burn_drive_get_media_size (NautilusBurnDrive *drive)
 	return size;
 }
 
+#ifdef USE_HAL
+static char *
+nautilus_burn_drive_hal_get_media_label (NautilusBurnDrive *drive)
+{
+	char          *label;
+	LibHalContext *ctx;
+	char          *child_udi;
+	DBusError      error;
+
+	g_return_val_if_fail (drive != NULL, NULL);
+
+	label = NULL;
+
+	if (drive->priv->udi == NULL) {
+		goto done;
+	}
+		
+	ctx = drive->priv->ctx;
+
+	if (ctx == NULL) {
+		goto done;
+	}
+
+	child_udi = hal_drive_get_first_child_udi (drive);
+	if (child_udi == NULL) {
+		goto done;
+	}
+
+	dbus_error_init (&error);
+	label = libhal_device_get_property_string (ctx, 
+						   child_udi,
+						   "volume.label",
+						   &error);
+	if (dbus_error_is_set (&error)) {
+		g_warning ("%s\n", error.message);
+		dbus_error_free (&error);
+	}
+
+	g_free (child_udi);
+ done:
+	return label;
+}
+#endif
+
+/**
+ * nautilus_burn_drive_get_media_label:
+ * @drive: #NautilusBurnDrive
+ *
+ * Determine the label of the media in the specified drive.
+ *
+ * Return value: The label of the media.
+ *
+ * Since: 2.14
+ **/
+char *
+nautilus_burn_drive_get_media_label (NautilusBurnDrive *drive)
+{
+	char *label;
+
+	g_return_val_if_fail (drive != NULL, NULL);
+
+	label = NULL;
+#ifdef USE_HAL
+	label = nautilus_burn_drive_hal_get_media_label (drive);
+#else
+	g_warning ("nautilus_burn_drive_get_media_label: not implemented without HAL");
+#endif
+
+	return label;
+}
+
 typedef struct {
 	gboolean    timeout;
 	gboolean    unmount_ok;
@@ -1338,6 +1541,8 @@ create_command (const char *device)
 		g_ptr_array_add (argv, str);
 		str = g_strdup ("--unmount");
 		g_ptr_array_add (argv, str);
+		str = g_strdup ("--no-ui");
+		g_ptr_array_add (argv, str);
 	} else {
 		char *str;
 		int   i;
@@ -1360,6 +1565,8 @@ create_command (const char *device)
 		g_ptr_array_add (argv, str);
 	}
 
+	g_ptr_array_add (argv, NULL);
+
 	return argv;
 }
 
@@ -1456,6 +1663,93 @@ nautilus_burn_drive_unmount (NautilusBurnDrive *drive)
 	return unmount_ok;
 }
 
+#ifdef USE_HAL
+static gboolean
+nautilus_burn_drive_hal_is_mounted (NautilusBurnDrive *drive)
+{
+	gboolean       is_mounted;
+	LibHalContext *ctx;
+	char          *child_udi;
+	DBusError      error;
+
+	g_return_val_if_fail (drive != NULL, FALSE);
+
+	is_mounted = FALSE;
+
+	if (drive->priv->udi == NULL) {
+		goto done;
+	}
+		
+	ctx = drive->priv->ctx;
+
+	if (ctx == NULL) {
+		goto done;
+	}
+
+	child_udi = hal_drive_get_first_child_udi (drive);
+	if (child_udi == NULL) {
+		goto done;
+	}
+
+	dbus_error_init (&error);
+	is_mounted = libhal_device_get_property_bool (ctx, 
+						      child_udi,
+						      "volume.is_mounted",
+						      &error);
+	if (dbus_error_is_set (&error)) {
+		g_warning ("%s\n", error.message);
+		dbus_error_free (&error);
+	}
+
+	g_free (child_udi);
+ done:
+	return is_mounted;
+}
+#else
+static gboolean
+nautilus_burn_drive_nohal_is_mounted (NautilusBurnDrive *drive)
+{
+	gboolean              is_mounted;
+	gint64                media_size;
+	NautilusBurnMediaType type;
+
+	media_size = nautilus_burn_drive_get_media_size (drive);
+	type = nautilus_burn_drive_get_media_type (drive);
+
+	is_mounted = (type == NAUTILUS_BURN_MEDIA_TYPE_BUSY)
+		|| (media_size == NAUTILUS_BURN_MEDIA_SIZE_BUSY);
+
+	return is_mounted;
+}
+#endif
+
+/**
+ * nautilus_burn_drive_is_mounted:
+ * @drive: #NautilusBurnDrive
+ *
+ * Determine if media in the specified drive is mounted.
+ *
+ * Return value: %TRUE if the media is mounted, %FALSE otherwise.
+ *
+ * Since: 2.14
+ **/
+gboolean
+nautilus_burn_drive_is_mounted (NautilusBurnDrive *drive)
+{
+	gboolean is_mounted;
+
+	g_return_val_if_fail (drive != NULL, FALSE);
+
+	is_mounted = FALSE;
+#ifdef USE_HAL
+	is_mounted = nautilus_burn_drive_hal_is_mounted (drive);
+#else
+	is_mounted = nautilus_burn_drive_nohal_is_mounted (drive);
+#endif
+
+	return is_mounted;
+}
+
 /**
  * nautilus_burn_drive_eject:
  * @drive: #NautilusBurnDrive
@@ -1478,7 +1772,7 @@ nautilus_burn_drive_eject (NautilusBurnDrive *drive)
 		return FALSE;
 	
 	if (use_gnome_mount ()) {
-		cmd = g_strdup_printf ("gnome-mount --eject --device=%s", drive->device);
+		cmd = g_strdup_printf ("gnome-mount --eject --no-ui --device=%s", drive->device);
 	} else {
 		cmd = g_strdup_printf ("eject %s", drive->device);
 	}
@@ -1621,6 +1915,41 @@ drive_hal_init (NautilusBurnDrive *drive)
 	drive->priv->ctx = ctx;
 }
 
+static int *
+hal_parse_write_speeds (char **strlist)
+{
+	char *end;
+	int  *write_speeds;
+	int   fields = 1;
+	int   i;
+
+	if (strlist == NULL)
+		return NULL;
+
+	fields = libhal_string_array_length (strlist);
+
+	if (fields == 0)
+		return NULL;
+
+	write_speeds = g_new0 (int, fields + 1);
+
+	for (i = 0; i < fields; i++) {
+		write_speeds[i] = strtol (strlist[i], &end, 10);
+
+		if (write_speeds[i] < 0
+		    || write_speeds[i] > 65535
+		    || *end != '\0') {
+
+			g_free (write_speeds);
+			return NULL;
+		}
+
+		write_speeds[i] /= CD_ROM_SPEED;
+	}
+
+	return write_speeds;
+}
+
 #define LIBHAL_PROP_EXTRACT_BEGIN if (FALSE)
 #define LIBHAL_PROP_EXTRACT_END ;
 #define LIBHAL_PROP_EXTRACT_INT(_property_, _where_) else if (strcmp (key, _property_) == 0 && type == LIBHAL_PROPERTY_TYPE_INT32) _where_ = libhal_psi_get_int (&it)
@@ -1628,7 +1957,7 @@ drive_hal_init (NautilusBurnDrive *drive)
 #define LIBHAL_PROP_EXTRACT_STRING(_property_, _where_) else if (strcmp (key, _property_) == 0 && type == LIBHAL_PROPERTY_TYPE_STRING) _where_ = (libhal_psi_get_string (&it) != NULL && strlen (libhal_psi_get_string (&it)) > 0) ? strdup (libhal_psi_get_string (&it)) : NULL
 #define LIBHAL_PROP_EXTRACT_BOOL(_property_, _where_) else if (strcmp (key, _property_) == 0 && type == LIBHAL_PROPERTY_TYPE_BOOLEAN) _where_ = libhal_psi_get_bool (&it)
 #define LIBHAL_PROP_EXTRACT_BOOL_BITFIELD(_property_, _where_, _field_) else if (strcmp (key, _property_) == 0 && type == LIBHAL_PROPERTY_TYPE_BOOLEAN) _where_ |= libhal_psi_get_bool (&it) ? _field_ : 0
-#define LIBHAL_PROP_EXTRACT_STRLIST(_property_, _where_) else if (strcmp (key, _property_) == 0 && type == LIBHAL_PROPERTY_TYPE_STRLIST) _where_ = my_strvdup (libhal_psi_get_strlist (&it))
+#define LIBHAL_PROP_EXTRACT_STRLIST(_property_, _where_) else if (strcmp (key, _property_) == 0 && type == LIBHAL_PROPERTY_TYPE_STRLIST) _where_ = libhal_psi_get_strlist (&it)
 
 static NautilusBurnDrive *
 hal_drive_from_udi (LibHalContext *ctx,
@@ -1638,6 +1967,7 @@ hal_drive_from_udi (LibHalContext *ctx,
 	LibHalPropertySetIterator it;
 	DBusError                 error;
 	NautilusBurnDrive        *drive;
+	char                    **write_speeds = NULL;
 
 	LIBHAL_CHECK_LIBHALCONTEXT (ctx, FALSE);
 
@@ -1664,12 +1994,12 @@ hal_drive_from_udi (LibHalContext *ctx,
 
 		LIBHAL_PROP_EXTRACT_BEGIN;
 
-		LIBHAL_PROP_EXTRACT_STRING  ("block.device",              drive->device);
-		LIBHAL_PROP_EXTRACT_STRING  ("storage.model",             drive->display_name);
-
-		LIBHAL_PROP_EXTRACT_INT_DIV ("storage.cdrom.read_speed",  drive->max_speed_read,  CD_ROM_SPEED);
-		LIBHAL_PROP_EXTRACT_INT_DIV ("storage.cdrom.write_speed", drive->max_speed_write, CD_ROM_SPEED);
+		LIBHAL_PROP_EXTRACT_STRING  ("block.device",               drive->device);
+		LIBHAL_PROP_EXTRACT_STRING  ("storage.model",              drive->display_name);
 
+		LIBHAL_PROP_EXTRACT_INT_DIV ("storage.cdrom.read_speed",   drive->max_speed_read,  CD_ROM_SPEED);
+		LIBHAL_PROP_EXTRACT_INT_DIV ("storage.cdrom.write_speed",  drive->max_speed_write, CD_ROM_SPEED);
+		LIBHAL_PROP_EXTRACT_STRLIST ("storage.cdrom.write_speeds", write_speeds);
 
 		LIBHAL_PROP_EXTRACT_BOOL_BITFIELD ("storage.cdrom.cdr",        drive->type, NAUTILUS_BURN_DRIVE_TYPE_CD_RECORDER);
 		LIBHAL_PROP_EXTRACT_BOOL_BITFIELD ("storage.cdrom.cdrw",       drive->type, NAUTILUS_BURN_DRIVE_TYPE_CDRW_RECORDER);
@@ -1684,14 +2014,19 @@ hal_drive_from_udi (LibHalContext *ctx,
 
 		LIBHAL_PROP_EXTRACT_END;
 	}
-
-	libhal_free_property_set (pset);
-
 	drive->priv->udi = g_strdup (udi);
 	drive->cdrecord_id = g_strdup (drive->device);
+	drive->priv->write_speeds = hal_parse_write_speeds (write_speeds);
+	/* we do not own write_speeds so do not free it. */
+	if (drive->priv->write_speeds == NULL) {
+		fill_write_speeds (drive);
+	}
 
-	if (! drive->display_name)
+	if (! drive->display_name) {
 		drive->display_name = g_strdup_printf ("Unnamed Drive (%s)", drive->device);
+	}
+
+	libhal_free_property_set (pset);
 
 	return drive;
 }
@@ -2063,6 +2398,7 @@ add_linux_cd_recorder (GList             *drives,
 		/* Have a wild guess, the drive should actually correct us */
 		drive->max_speed_write = drive_s->speed;
 	}
+	fill_write_speeds (drive);
 
 	drive->device = g_strdup_printf ("/dev/%s", drive_s->device);
 	drive->max_speed_read = drive_s->speed;
@@ -2102,6 +2438,7 @@ add_linux_cd_drive (GList             *drives,
 	drive->display_name = g_strdup (drive_s->display_name);
 	drive->device = g_strdup_printf ("/dev/%s", drive_s->device);
 	drive->max_speed_write = 0; /* Can't write */
+	fill_write_speeds (drive);
 	drive->max_speed_read = drive_s->speed;
 	if (drive_s->can_read_dvd) {
 		drive->type |= NAUTILUS_BURN_DRIVE_TYPE_DVD_DRIVE;
@@ -2400,6 +2737,7 @@ freebsd_scan (gboolean recorder_only)
 			if (drive->max_speed_write == -1) {
 			    	drive->max_speed_write = speed;
 			}
+			fill_write_speeds (drive);
 
 			if (drive->type & NAUTILUS_BURN_DRIVE_TYPE_DVD_DRIVE) {
 				add_dvd_plus (drive);
@@ -2503,6 +2841,7 @@ cdrecord_scan (gboolean recorder_only)
 			if (drive->max_speed_write == -1) {
 			    	drive->max_speed_write = DEFAULT_SPEED;
 			}
+			fill_write_speeds (drive);
 
 			if (drive->type & NAUTILUS_BURN_DRIVE_TYPE_DVD_DRIVE) {
 				add_dvd_plus (drive);
@@ -2537,6 +2876,7 @@ nautilus_burn_drive_get_file_image (void)
 	drive->display_name = g_strdup (_("File image"));
 	drive->max_speed_read = 0;
 	drive->max_speed_write = 0;
+	fill_write_speeds (drive);
 	drive->type = NAUTILUS_BURN_DRIVE_TYPE_FILE;
 
 	return drive;
@@ -2763,6 +3103,7 @@ nautilus_burn_drive_finalize (GObject *object)
 	g_free (drive->priv->udi);
 	g_free (drive->priv->monitor_udi);
 
+	g_free (drive->priv->write_speeds);
 	g_free (drive->display_name);
 	g_free (drive->cdrecord_id);
 	g_free (drive->device);
@@ -2781,6 +3122,7 @@ nautilus_burn_drive_init (NautilusBurnDrive *drive)
 	drive->display_name    = NULL;
 	drive->max_speed_write = 0;
 	drive->max_speed_read  = 0;
+	drive->priv->write_speeds = NULL;
 	drive->cdrecord_id     = NULL;
 	drive->device          = NULL;
 
@@ -2986,6 +3328,58 @@ nautilus_burn_drive_media_type_get_string (NautilusBurnMediaType type)
 	return _("Broken media type");
 }
 
+/**
+ * nautilus_burn_drive_media_type_is_writable
+ * @type: A #NautilusBurnMediaType
+ * @is_blank: if the media type is blank or not
+ *
+ * Determine if a media type is writable
+ *
+ * Returns: %TRUE if the media type can be written to, otherwise return %FALSE.
+ *
+ * Since: 2.14
+ **/
+gboolean
+nautilus_burn_drive_media_type_is_writable (NautilusBurnMediaType type,
+					    gboolean              is_blank)
+{
+	gboolean can_write;
+
+	can_write = FALSE;
+
+	switch (type) {
+	case NAUTILUS_BURN_MEDIA_TYPE_BUSY:
+	case NAUTILUS_BURN_MEDIA_TYPE_ERROR:
+	case NAUTILUS_BURN_MEDIA_TYPE_UNKNOWN:
+	case NAUTILUS_BURN_MEDIA_TYPE_CD:
+	case NAUTILUS_BURN_MEDIA_TYPE_DVD:
+	case NAUTILUS_BURN_MEDIA_TYPE_DVD_RAM:
+		can_write = FALSE;
+		break;
+	case NAUTILUS_BURN_MEDIA_TYPE_CDR:
+	case NAUTILUS_BURN_MEDIA_TYPE_DVDR:
+	case NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_R:
+	case NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_R_DL:
+		if (is_blank) {
+			can_write = TRUE;
+		} else {
+			can_write = FALSE;
+		}
+		break;
+	case NAUTILUS_BURN_MEDIA_TYPE_CDRW:
+	case NAUTILUS_BURN_MEDIA_TYPE_DVDRW:
+	case NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_RW:
+		can_write = TRUE;
+		break;
+	default:
+		g_warning ("Unknown media type: %d", type);
+		can_write = FALSE;
+		break;
+	}
+
+	return can_write;
+}
+
 #ifdef USE_HAL
 static void
 set_monitor_enabled_hal (NautilusBurnDrive *drive,
@@ -2996,31 +3390,16 @@ set_monitor_enabled_hal (NautilusBurnDrive *drive,
 	drive->priv->monitor_udi = NULL;
 
 	if (enabled) {
-		char    **device_names;
-		int       num_devices = -1;
-		DBusError error;
+		char *child_udi;
 
 		libhal_ctx_set_user_data (drive->priv->ctx, drive);
 		libhal_ctx_set_device_added (drive->priv->ctx, hal_device_added);
 		libhal_ctx_set_device_removed (drive->priv->ctx, hal_device_removed);
 
-		dbus_error_init (&error);
-		device_names = libhal_manager_find_device_string_match (drive->priv->ctx, 
-									"info.parent",
-									drive->priv->udi,
-									&num_devices,
-									&error);
-		
-		if (dbus_error_is_set (&error)) {
-			g_warning ("%s\n", error.message);
-			dbus_error_free (&error);
+		child_udi = hal_drive_get_first_child_udi (drive);
+		if (child_udi != NULL) {
+			drive->priv->monitor_udi = child_udi;
 		}
-
-		if (num_devices > 0) {
-			drive->priv->monitor_udi = g_strdup (device_names [0]);
-		}
-
-		libhal_free_string_array (device_names);
 	} else {
 		libhal_ctx_set_user_data (drive->priv->ctx, NULL);
 		libhal_ctx_set_device_added (drive->priv->ctx, NULL);
